//////////////////////////////////////////////////////
// Systems Development Lab., Hitachi, Ltd.
// Low Power Consumption Random Number Generator
// enocoro-128 v2
// reference hardware code Rev.1.0
// 
// 2 February 2010
// (c) Hitachi, Ltd. 2010. All rights reserved.
//////////////////////////////////////////////////////

module enocoro_128 (
   reset_n
  ,clk

  ,key_in
  ,iv_in
  ,input_set
  ,output_ready
  ,output_start
  ,random_out

);

  input          reset_n;
  input          clk;

  input  [127:0] key_in;
  input  [ 63:0] iv_in;
  input          input_set;
  input          output_start;

  output         output_ready;
  output [  7:0] random_out;

`define ib24 8'h66
`define ib25 8'he9
`define ib26 8'h4b
`define ib27 8'hd4
`define ib28 8'hef
`define ib29 8'h8a
`define ib30 8'h2c
`define ib31 8'h3b

`define ia0 8'h88
`define ia1 8'h4c

`define ictr 8'h01


`define L 1'b0
`define H 1'b1

  reg  [7:0] b00,b01,b02,b03,b04,b05,b06,b07
            ,b08,b09,b10,b11,b12,b13,b14,b15
            ,b16,b17,b18,b19,b20,b21,b22,b23
            ,b24,b25,b26,b27,b28,b29,b30,b31;
  reg  [7:0] a0,a1;

  wire       dout;
  reg  [6:0] count;
  reg  [1:0] stage;

  reg  [7:0] ctr;

// b with  function

  always@(posedge clk or negedge reset_n)
  if(reset_n == `L) begin
    b00 <= 8'h00; b01 <= 8'h00; b02 <= 8'h00; b03 <= 8'h00; b04 <= 8'h00; b05 <= 8'h00; b06 <= 8'h00; b07 <= 8'h00;
    b08 <= 8'h00; b09 <= 8'h00; b10 <= 8'h00; b11 <= 8'h00; b12 <= 8'h00; b13 <= 8'h00; b14 <= 8'h00; b15 <= 8'h00;
    b16 <= 8'h00; b17 <= 8'h00; b18 <= 8'h00; b19 <= 8'h00; b20 <= 8'h00; b21 <= 8'h00; b22 <= 8'h00; b23 <= 8'h00;
    b24 <= 8'h00; b25 <= 8'h00; b26 <= 8'h00; b27 <= 8'h00; b28 <= 8'h00; b29 <= 8'h00; b30 <= 8'h00; b31 <= 8'h00; end
  else
    if(input_set == `H) begin
      b00 <= key_in[127:120]; b01 <= key_in[119:112]; b02 <= key_in[111:104]; b03 <= key_in[103: 96];
      b04 <= key_in[ 95: 88]; b05 <= key_in[ 87: 80]; b06 <= key_in[ 79: 72]; b07 <= key_in[ 71: 64];
      b08 <= key_in[ 63: 56]; b09 <= key_in[ 55: 48]; b10 <= key_in[ 47: 40]; b11 <= key_in[ 39: 32];
      b12 <= key_in[ 31: 24]; b13 <= key_in[ 23: 16]; b14 <= key_in[ 15:  8]; b15 <= key_in[  7:  0];
      b16 <=  iv_in[ 63: 56]; b17 <=  iv_in[ 55: 48]; b18 <=  iv_in[ 47: 40]; b19 <=  iv_in[ 39: 32];
      b20 <=  iv_in[ 31: 24]; b21 <=  iv_in[ 23: 16]; b22 <=  iv_in[ 15:  8]; b23 <=  iv_in[  7:  0];
      b24 <= `ib24;        b25 <= `ib25;        b26 <= `ib26;        b27 <= `ib27;
      b28 <= `ib28;        b29 <= `ib29;        b30 <= `ib30;        b31 <= `ib31;        end
    else
      if(dout == `H) begin
        b00 <= b31 ^ a0 ^ ctr;
        b01 <= b00;       b02 <= b01;       b03 <= b02 ^ b06;
        b04 <= b03;       b05 <= b04;       b06 <= b05;       b07 <= b06;
        b08 <= b07 ^ b15; b09 <= b08;       b10 <= b09;       b11 <= b10;
        b12 <= b11;       b13 <= b12;       b14 <= b13;       b15 <= b14;
        b16 <= b15;       b17 <= b16 ^ b28; b18 <= b17;       b19 <= b18;
        b20 <= b19;       b21 <= b20;       b22 <= b21;       b23 <= b22;
        b24 <= b23;       b25 <= b24;       b26 <= b25;       b27 <= b26;
        b28 <= b27;       b29 <= b28;       b30 <= b29;       b31 <= b30;       
        end
      else begin
        b00 <= b00; b01 <= b01; b02 <= b02; b03 <= b03; b04 <= b04; b05 <= b05; b06 <= b06; b07 <= b07;
        b08 <= b08; b09 <= b09; b10 <= b10; b11 <= b11; b12 <= b12; b13 <= b13; b14 <= b14; b15 <= b15;
        b16 <= b16; b17 <= b17; b18 <= b18; b19 <= b19; b20 <= b20; b21 <= b21; b22 <= b22; b23 <= b23;
        b24 <= b24; b25 <= b25; b26 <= b26; b27 <= b27; b28 <= b28; b29 <= b29; b30 <= b30; b31 <= b31; 
      end

// a

  wire [15:0] rho_out;
    assign rho_out = rho(a0,a1,b02,b07,b16,b29);

  always@(posedge clk or negedge reset_n)
  if(reset_n == `L) begin
    a0 <= 8'h00; a1 <= 8'h00; end
  else
    if(input_set == `H) begin
      a0 <= `ia0; a1 <= `ia1; end
    else
      if(dout == `H) begin
        a0 <= rho_out[15:8]; 
        a1 <= rho_out[ 7:0]; end
      else begin
        a0 <= a0; a1 <= a1; end

// ctr

  always    @(posedge clk or negedge reset_n)
  begin
      if(reset_n == 1'b0)begin
          ctr   <=  8'b00;
      end
      else if(input_set == `H) begin
          ctr   <=  `ictr;
      end
      else if(count == 7'h00)begin
          ctr   <=  8'h00;
      end
      else if(output_ready == `L)begin
          ctr   <=  gf_8_x_alpha(ctr);
      end
  end

// enocoro main stage

`define enocoro_idle   2'b00
`define enocoro_key_iv 2'b01
`define enocoro_ready  2'b11

  always@(posedge clk or negedge reset_n)
  if(reset_n == `L)        stage <= `enocoro_idle;
  else
    case(stage)
      `enocoro_idle   :
        if(input_set == `H)      stage <= `enocoro_key_iv;
        else               stage <= stage;
      `enocoro_key_iv :
        if(count == 7'h00) stage <= `enocoro_ready;
        else               stage <= stage;
      `enocoro_ready  :
        if(input_set == `H)      stage <= `enocoro_key_iv;
        else               stage <= stage;
      default:             stage <= `enocoro_idle;
    endcase

  always@(posedge clk or negedge reset_n)
  if(reset_n == `L)      count <= 7'h00;
  else
    if(input_set == `H)        count <= 7'h5f;
    else
      if(count == 7'h00) count <= count;
      else               count <= count - 7'h01;

  assign dout = (stage == `enocoro_key_iv)
            | (stage == `enocoro_ready) & output_start;

  assign output_ready = (stage == `enocoro_ready);

// enocoro output

  assign random_out = a1;

function [15:0] rho;
input    [ 7:0] a0;
input    [ 7:0] a1;
input    [ 7:0] k1;
input    [ 7:0] k2;
input    [ 7:0] k3;
input    [ 7:0] k4;
reg      [15:0] temp;

begin

  temp[15:0] = Liner({(a0 ^ s8(k1)),(a1 ^ s8(k2))});
  rho = {(temp[15:8] ^ s8(k3)),(temp[7:0] ^ s8(k4))};

end
endfunction

function [15:0] Liner;
input    [15:0] in;

  Liner = {(in[15:8] ^ in[ 7:0]),(in[15:8] ^ gf_8_x_alpha(in[ 7:0]))};

endfunction

function [7:0] s8;
input   [7:0]  in;
begin
    case(in)
        8'h00: s8 = 8'h63; 8'h01: s8 = 8'h52; 8'h02: s8 = 8'h1a; 8'h03: s8 = 8'hdf;
        8'h04: s8 = 8'h8a; 8'h05: s8 = 8'hf6; 8'h06: s8 = 8'hae; 8'h07: s8 = 8'h55;
        8'h08: s8 = 8'h89; 8'h09: s8 = 8'he7; 8'h0a: s8 = 8'hd0; 8'h0b: s8 = 8'h2d;
        8'h0c: s8 = 8'hbd; 8'h0d: s8 = 8'h01; 8'h0e: s8 = 8'h24; 8'h0f: s8 = 8'h78;
        8'h10: s8 = 8'h1b; 8'h11: s8 = 8'hd9; 8'h12: s8 = 8'he3; 8'h13: s8 = 8'h54;
        8'h14: s8 = 8'hc8; 8'h15: s8 = 8'ha4; 8'h16: s8 = 8'hec; 8'h17: s8 = 8'h7e;
        8'h18: s8 = 8'hab; 8'h19: s8 = 8'h00; 8'h1a: s8 = 8'h9c; 8'h1b: s8 = 8'h2e;
        8'h1c: s8 = 8'h91; 8'h1d: s8 = 8'h67; 8'h1e: s8 = 8'h37; 8'h1f: s8 = 8'h53;
        8'h20: s8 = 8'h4e; 8'h21: s8 = 8'h6b; 8'h22: s8 = 8'h6c; 8'h23: s8 = 8'h11;
        8'h24: s8 = 8'hb2; 8'h25: s8 = 8'hc0; 8'h26: s8 = 8'h82; 8'h27: s8 = 8'hfd;
        8'h28: s8 = 8'h39; 8'h29: s8 = 8'h45; 8'h2a: s8 = 8'hfe; 8'h2b: s8 = 8'h9b;
        8'h2c: s8 = 8'h34; 8'h2d: s8 = 8'hd7; 8'h2e: s8 = 8'ha7; 8'h2f: s8 = 8'h08;
        8'h30: s8 = 8'hb8; 8'h31: s8 = 8'h9a; 8'h32: s8 = 8'h33; 8'h33: s8 = 8'hc6;
        8'h34: s8 = 8'h4c; 8'h35: s8 = 8'h1d; 8'h36: s8 = 8'h69; 8'h37: s8 = 8'ha1;
        8'h38: s8 = 8'h6e; 8'h39: s8 = 8'h3e; 8'h3a: s8 = 8'hc5; 8'h3b: s8 = 8'h0a;
        8'h3c: s8 = 8'h57; 8'h3d: s8 = 8'hf4; 8'h3e: s8 = 8'hf1; 8'h3f: s8 = 8'h83;
        8'h40: s8 = 8'hf5; 8'h41: s8 = 8'h47; 8'h42: s8 = 8'h1f; 8'h43: s8 = 8'h7a;
        8'h44: s8 = 8'ha5; 8'h45: s8 = 8'h29; 8'h46: s8 = 8'h3c; 8'h47: s8 = 8'h42;
        8'h48: s8 = 8'hd6; 8'h49: s8 = 8'h73; 8'h4a: s8 = 8'h8d; 8'h4b: s8 = 8'hf0;
        8'h4c: s8 = 8'h8e; 8'h4d: s8 = 8'h18; 8'h4e: s8 = 8'haa; 8'h4f: s8 = 8'hc1;
        8'h50: s8 = 8'h20; 8'h51: s8 = 8'hbf; 8'h52: s8 = 8'he6; 8'h53: s8 = 8'h93;
        8'h54: s8 = 8'h51; 8'h55: s8 = 8'h0e; 8'h56: s8 = 8'hf7; 8'h57: s8 = 8'h98;
        8'h58: s8 = 8'hdd; 8'h59: s8 = 8'hba; 8'h5a: s8 = 8'h6a; 8'h5b: s8 = 8'h05;
        8'h5c: s8 = 8'h48; 8'h5d: s8 = 8'h23; 8'h5e: s8 = 8'h6d; 8'h5f: s8 = 8'hd4;
        8'h60: s8 = 8'h1e; 8'h61: s8 = 8'h60; 8'h62: s8 = 8'h75; 8'h63: s8 = 8'h43;
        8'h64: s8 = 8'h97; 8'h65: s8 = 8'h2a; 8'h66: s8 = 8'h31; 8'h67: s8 = 8'hdb;
        8'h68: s8 = 8'h84; 8'h69: s8 = 8'h19; 8'h6a: s8 = 8'haf; 8'h6b: s8 = 8'hbc;
        8'h6c: s8 = 8'hcc; 8'h6d: s8 = 8'hf3; 8'h6e: s8 = 8'he8; 8'h6f: s8 = 8'h46;
        8'h70: s8 = 8'h88; 8'h71: s8 = 8'hac; 8'h72: s8 = 8'h8b; 8'h73: s8 = 8'he4;
        8'h74: s8 = 8'h7b; 8'h75: s8 = 8'hd5; 8'h76: s8 = 8'h58; 8'h77: s8 = 8'h36;
        8'h78: s8 = 8'h02; 8'h79: s8 = 8'hb1; 8'h7a: s8 = 8'h07; 8'h7b: s8 = 8'h72;
        8'h7c: s8 = 8'he1; 8'h7d: s8 = 8'hdc; 8'h7e: s8 = 8'h5f; 8'h7f: s8 = 8'h2f;
        8'h80: s8 = 8'h5d; 8'h81: s8 = 8'he5; 8'h82: s8 = 8'hd1; 8'h83: s8 = 8'h0c;
        8'h84: s8 = 8'h26; 8'h85: s8 = 8'h99; 8'h86: s8 = 8'hb5; 8'h87: s8 = 8'h6f;
        8'h88: s8 = 8'he0; 8'h89: s8 = 8'h4a; 8'h8a: s8 = 8'h3b; 8'h8b: s8 = 8'hde;
        8'h8c: s8 = 8'ha2; 8'h8d: s8 = 8'h68; 8'h8e: s8 = 8'h92; 8'h8f: s8 = 8'h17;
        8'h90: s8 = 8'hca; 8'h91: s8 = 8'hee; 8'h92: s8 = 8'ha9; 8'h93: s8 = 8'hb6;
        8'h94: s8 = 8'h03; 8'h95: s8 = 8'h5e; 8'h96: s8 = 8'hd3; 8'h97: s8 = 8'h25;
        8'h98: s8 = 8'hfb; 8'h99: s8 = 8'h9d; 8'h9a: s8 = 8'h61; 8'h9b: s8 = 8'h59;
        8'h9c: s8 = 8'h06; 8'h9d: s8 = 8'h90; 8'h9e: s8 = 8'h74; 8'h9f: s8 = 8'h2c;
        8'ha0: s8 = 8'h27; 8'ha1: s8 = 8'h95; 8'ha2: s8 = 8'ha0; 8'ha3: s8 = 8'hb9;
        8'ha4: s8 = 8'h7c; 8'ha5: s8 = 8'hed; 8'ha6: s8 = 8'h04; 8'ha7: s8 = 8'hd2;
        8'ha8: s8 = 8'h50; 8'ha9: s8 = 8'he2; 8'haa: s8 = 8'h49; 8'hab: s8 = 8'h77;
        8'hac: s8 = 8'hcb; 8'had: s8 = 8'h3a; 8'hae: s8 = 8'h0f; 8'haf: s8 = 8'h9e;
        8'hb0: s8 = 8'h70; 8'hb1: s8 = 8'h16; 8'hb2: s8 = 8'h5c; 8'hb3: s8 = 8'hef;
        8'hb4: s8 = 8'h21; 8'hb5: s8 = 8'hb3; 8'hb6: s8 = 8'h9f; 8'hb7: s8 = 8'h0d;
        8'hb8: s8 = 8'ha6; 8'hb9: s8 = 8'hc9; 8'hba: s8 = 8'h22; 8'hbb: s8 = 8'h94;
        8'hbc: s8 = 8'hfa; 8'hbd: s8 = 8'h4b; 8'hbe: s8 = 8'hd8; 8'hbf: s8 = 8'h65;
        8'hc0: s8 = 8'h85; 8'hc1: s8 = 8'h3d; 8'hc2: s8 = 8'h96; 8'hc3: s8 = 8'h28;
        8'hc4: s8 = 8'h14; 8'hc5: s8 = 8'h5b; 8'hc6: s8 = 8'h66; 8'hc7: s8 = 8'hea;
        8'hc8: s8 = 8'h7f; 8'hc9: s8 = 8'hce; 8'hca: s8 = 8'hf9; 8'hcb: s8 = 8'h40;
        8'hcc: s8 = 8'h13; 8'hcd: s8 = 8'had; 8'hce: s8 = 8'hc3; 8'hcf: s8 = 8'hb0;
        8'hd0: s8 = 8'hf2; 8'hd1: s8 = 8'hc2; 8'hd2: s8 = 8'h38; 8'hd3: s8 = 8'h80;
        8'hd4: s8 = 8'hcf; 8'hd5: s8 = 8'h71; 8'hd6: s8 = 8'h0b; 8'hd7: s8 = 8'h87;
        8'hd8: s8 = 8'h4d; 8'hd9: s8 = 8'h35; 8'hda: s8 = 8'h56; 8'hdb: s8 = 8'he9;
        8'hdc: s8 = 8'h64; 8'hdd: s8 = 8'hbe; 8'hde: s8 = 8'h1c; 8'hdf: s8 = 8'hbb;
        8'he0: s8 = 8'hb7; 8'he1: s8 = 8'h30; 8'he2: s8 = 8'hc4; 8'he3: s8 = 8'h2b;
        8'he4: s8 = 8'hff; 8'he5: s8 = 8'h62; 8'he6: s8 = 8'h41; 8'he7: s8 = 8'ha8;
        8'he8: s8 = 8'h15; 8'he9: s8 = 8'h8c; 8'hea: s8 = 8'h12; 8'heb: s8 = 8'hc7;
        8'hec: s8 = 8'h79; 8'hed: s8 = 8'h8f; 8'hee: s8 = 8'h5a; 8'hef: s8 = 8'hfc;
        8'hf0: s8 = 8'hcd; 8'hf1: s8 = 8'h09; 8'hf2: s8 = 8'h4f; 8'hf3: s8 = 8'h7d;
        8'hf4: s8 = 8'hf8; 8'hf5: s8 = 8'h86; 8'hf6: s8 = 8'hda; 8'hf7: s8 = 8'h10;
        8'hf8: s8 = 8'h32; 8'hf9: s8 = 8'h76; 8'hfa: s8 = 8'hb4; 8'hfb: s8 = 8'ha3;
        8'hfc: s8 = 8'h3f; 8'hfd: s8 = 8'h44; 8'hfe: s8 = 8'h81; 8'hff: s8 = 8'heb;
        default: s8[3:0] = 8'hxx;
    endcase
end
endfunction


function [7:0] gf_8_x_alpha ;
input    [7:0] in ;

// primitive function x^8 + x^4 + x^3 + x^2 + 1
  gf_8_x_alpha = {in[6],in[5],in[4],(in[3]^in[7]),(in[2]^in[7]),(in[1]^in[7]),in[0],in[7]};

endfunction

function [3:0] gf_4_x_alpha ;
input    [3:0] in ;

// primitive function x^4 + x +1
  gf_4_x_alpha = {in[2],in[1],(in[0]^in[3]),in[3]};

endfunction

endmodule
